import requests
from unittest.mock import patch
from datetime import date, timedelta
from dateutil.relativedelta import relativedelta
from freezegun import freeze_time

from odoo import fields
import odoo.tools as tools
from odoo.tests.common import HttpCase, tagged, new_test_user
from odoo.addons.openerp_enterprise.tests.common import EnterpriseDatabaseCommon, generate_uuid, now


class DatabaseTestCase(EnterpriseDatabaseCommon):
    def test_create_database(self):
        record_id = self.create_database(generate_uuid())

        self.assertIsNotNone(record_id)


class PingTestCase(EnterpriseDatabaseCommon):

    def _user_ping_info(self, user=None, login=False, name=False, admin=False):
        userinfo = {
            'name': user.name if user else name,
            'login': user.login if user else login,
            'is_admin': admin,
            'login_date': fields.Datetime.to_string(fields.Datetime.now()),
        }
        if user:
            userinfo.update(odoo_com_uid=user.id, oauth=True)
        return userinfo

    @patch('odoo.addons.openerp_enterprise.models.openerp_enterprise.ip2coordinates')
    def test_create_ping(self, ip2coordinates_mock):
        ip2coordinates_mock.return_value = {'country_code': 'GB', 'city': 'Unlocalized'}
        database = self.create_database()
        values = {
            'database': database.id,
            'nbr_active_users': 0,
            'nbr_share_users': 0,
            'nbr_active_share_users': 0,
            'nbr_users': 0,
            'version': 6.1,
            'ping_time': now(),
            'ip': '217.194.208.100',
            'language': 'fr_BE',
            'company_name': 'OpenERP',
            'company_phone': '+32 81 81 37 00',
            'company_email': 'info@openerp.com',
            'db_name': 'test',
        }
        ping = self.env['openerp.enterprise.ping'].create(values)
        self.assertIsNotNone(ping)
        self.assertEqual(ping.country.name, 'United Kingdom')

    def test_last_ping(self):
        database = self.create_database()
        values = {
            'database': database.id,
            'nbr_active_users': 0,
            'nbr_share_users': 0,
            'nbr_active_share_users': 0,
            'nbr_users': 0,
            'version': 6.1,
            'ping_time': now(),
            'ip': '127.0.0.1',
            'language': 'fr_BE',
            'company_name': 'OpenERP',
            'company_phone': '+32 81 81 37 00',
            'company_email': 'info@openerp.com',
            'db_name': 'test',
        }

        ping = self.env['openerp.enterprise.ping'].create(values)
        self.assertIsNotNone(ping)

        self.assertEqual(database.first_ping, date.today())
        self.assertEqual(database.last_ping, date.today())
        self.assertEqual(ping.ping_time.date(), database.last_ping)

    def test_last_ping_multi(self):
        proxy = self.env['openerp.enterprise.ping']

        database = self.create_database()

        current_datetime = date.today()
        first_ping = date.today()

        for i in range(1, 10):
            values = {
                'database': database.id,
                'nbr_active_users': 0,
                'nbr_share_users': 0,
                'nbr_active_share_users': 0,
                'nbr_users': 0,
                'version': 6.1,
                'ping_time': current_datetime,
                'ip': '127.0.0.1',
                'language': 'fr_BE',
                'company_name': 'OpenERP',
                'company_phone': '+32 81 81 37 00',
                'company_email': 'info@openerp.com',
                'db_name': 'test',
            }

            record_id = proxy.create(values)
            self.assertIsNotNone(record_id)

            current_datetime = current_datetime + relativedelta(days=1)

        last_ping = first_ping + relativedelta(days=len(range(1, 9)))

        self.assertEqual(database.first_ping, first_ping)
        self.assertEqual(database.last_ping, last_ping)

    def test_ping_user_oauth_link(self):
        # test that database's user link with odoo.com accounts are correctly
        # updated with ping
        db_user1 = new_test_user(
            self.env, login='user1@db.example.com', name="User 1",
            groups="base.group_portal",
        )
        db_user2 = new_test_user(
            self.env, login='user2@db.example.com', name="User 2",
            groups="base.group_portal",
        )

        database = self.create_database()
        self.assertEqual(database.quantity, 1, "Quantity should be 1 when no ping data available")

        # send ping with users linked to their respective odoo.com account
        with freeze_time("2023-01-16"):
            self.send_ping(database.db_uuid, {}, payload_update={
                'nbr_users': 2,
                'nbr_active_users': 2,
                'users': [
                    self._user_ping_info(db_user1, admin=True),
                    self._user_ping_info(db_user2),
                ]
            })
        self.assertEqual(database.nbr_active_regular_users, 2)
        self.assertEqual(database.quantity, 2)
        self.assertRecordValues(
            database.registered_user_ids,
            [
                {'user_id': db_user1.id, 'is_admin': True},
                {'user_id': db_user2.id, 'is_admin': False},
            ]
        )
        self.assertRecordValues(
            database.database_user_ids,
            [
                {'name': db_user1.name, 'login': db_user1.login, 'user_id': db_user1.id},
                {'name': db_user2.name, 'login': db_user2.login, 'user_id': db_user2.id},
            ]
        )

        # simulate "oauth_admin_change" that unlink db_user2 from it odoo.com account
        # - db_user2 should be removed from the oauth user list
        # - db_user2 should stay listed as database user but without a odoo.com user_id
        with freeze_time("2023-01-23"):
            self.send_ping(database.db_uuid, {}, payload_update={
                'nbr_users': 2,
                'nbr_active_users': 2,
                'users': [
                    self._user_ping_info(db_user1, admin=True),
                    self._user_ping_info(login=db_user2.login, name=db_user2.name),  # no oauth user
                ]
            })
        self.assertEqual(database.nbr_active_regular_users, 2)
        self.assertEqual(database.quantity, 2)
        self.assertRecordValues(
            database.registered_user_ids,
            [
                {'user_id': db_user1.id, 'is_admin': True},
            ]
        )
        self.assertRecordValues(
            database.database_user_ids,
            [
                {'name': db_user1.name, 'login': db_user1.login, 'user_id': db_user1.id},
                {'name': db_user2.name, 'login': db_user2.login, 'user_id': False},  # no oauth user
            ]
        )

    def test_ping_free_accounting_firm_user(self):
        """ Ensure that we do not count database users linked to an account firm as  billable. """
        accounting_firm_partner = self.env["res.partner"].create({
            'name': 'My Accounting Firm',
            'portal_accounting_firm_level': '2_silver',
            'is_company': True,
            'is_published': True,
        })
        acct1_user = new_test_user(
            self.env, login='acct1@acct-firm.example.com', name="Accountant 1",
            groups="base.group_portal", parent_id=accounting_firm_partner.id,
        )
        acct2_user = new_test_user(
            self.env, login='acct2@acct-firm.example.com', name="Accoutant 2",
            groups="base.group_portal", parent_id=accounting_firm_partner.id,
        )

        database = self.create_database()
        self.assertEqual(database.quantity, 1, "Quantity should be 1 when no ping data available")

        # 1. Send ping: all users are linked to an accounting firm
        with freeze_time("2023-01-16"):
            self.send_ping(database.db_uuid, {}, payload_update={
                'nbr_users': 2,
                'nbr_active_users': 2,
                'users': [
                    self._user_ping_info(acct1_user),
                    self._user_ping_info(acct2_user),
                ]
            })
        self.assertEqual(database.nbr_active_regular_users, 2)
        self.assertEqual(database.quantity, 1, "we should have a minimal regular user count of 1")

        # 2. End users decide to use their instance, so new users have been created
        # but only locally
        with freeze_time("2023-01-23"):
            self.send_ping(database.db_uuid, {}, payload_update={
                'nbr_users': 4,
                'nbr_active_users': 4,
                'users': [
                    self._user_ping_info(acct1_user, admin=True),
                    self._user_ping_info(acct2_user),
                    self._user_ping_info(name="Customer 1", login="customer1@my.example.com", admin=True),
                    self._user_ping_info(name="Customer 2", login="customer2@my.example.com"),
                ]
            })
        self.assertEqual(database.nbr_active_regular_users, 4)
        self.assertEqual(database.quantity, 2)

        # 2. One user decides to create and link it's account on odoo.com
        cust1_user = new_test_user(
            self.env, login='customer1@my.example.com', name="Customer 1",
            groups="base.group_portal",
        )

        with freeze_time("2023-01-24"):
            self.send_ping(database.db_uuid, {}, payload_update={
                'nbr_users': 4,
                'nbr_active_users': 4,
                'users': [
                    self._user_ping_info(acct1_user, admin=True),
                    self._user_ping_info(acct2_user),
                    self._user_ping_info(cust1_user, admin=True),
                    self._user_ping_info(name="Customer 2", login="customer2@my.example.com"),
                ]
            })
        self.assertEqual(database.nbr_active_regular_users, 4)
        self.assertEqual(database.quantity, 2)
        self.assertEqual(database.last_ping_id.nbr_users, 4)
        self.assertEqual(database.last_ping_id.nbr_active_users, 4)

        # 3. If accounting firms then decided to not be referenced anymore
        # on the website, their users should be counted as normal user.
        accounting_firm_partner.is_published = False
        with freeze_time("2023-02-06"):
            self.send_ping(database.db_uuid, {}, payload_update={
                'nbr_users': 4,
                'nbr_active_users': 4,
                'users': [
                    self._user_ping_info(acct1_user, admin=True),
                    self._user_ping_info(acct2_user),
                    self._user_ping_info(cust1_user, admin=True),
                    self._user_ping_info(name="Customer 2", login="customer2@my.example.com"),
                ]
            })
        self.assertEqual(database.nbr_active_regular_users, 4)
        self.assertEqual(database.quantity, 4)
        self.assertEqual(database.last_ping_id.nbr_users, 4)
        self.assertEqual(database.last_ping_id.nbr_active_users, 4)

@tagged('post_install', '-at_install')
class HTTPPostUnitTest(EnterpriseDatabaseCommon, HttpCase):

    def setUp(self):
        super().setUp()
        force_ping_db_patcher = patch(
            'odoo.addons.openerp_enterprise.controllers.main.PUBLISHER_WARRANTY_DB',
            new=self.env.cr.dbname
        )
        self.startPatcher(force_ping_db_patcher)

    def test_external_ping_old_style(self):
        # publisher_warranty_url = "http://services.openerp.com/publisher-warranty/"
        publisher_warranty_url = self.base_url() + '/publisher-warranty/'
        dbuuid = "21d20d8a-03d8-11e2-86c4-c4461903c064"
        # dbuuid = "21d20d8a-03d8-11e2-86c4-c4461903c065"
        db_create_date = '2012-09-21 10:00:00'

        values = {
            'dbuuid': dbuuid,
            'nbr_users': 0,
            'nbr_active_users': 0,
            'nbr_share_users': 0,
            'nbr_active_share_users': 0,
            'dbname': 'stw_dontuse_test',
            'db_create_date': db_create_date,
            'version': '7.0-alpha2',
            'contracts': [],
            'language': 'en_US',
            'name': 'Stephane Wirtel',
            'phone': '+',
            'email': 'stw@openerp.com',

        }

        params = {
            'arg0': repr(values),
            'action': 'update',
        }

        query = requests.get(publisher_warranty_url, params=params)  # , config={'verbose' : sys.stderr})

        self.assertEqual(query.status_code, 200)

    def test_external_ping_new_style(self):
        # publisher_warranty_url = "http://services.openerp.com/publisher-warranty/"
        publisher_warranty_url = self.base_url() + '/publisher-warranty/'
        dbuuid = "21d20d8a-03d8-11e2-86c4-c4461903c064"
        # dbuuid = "21d20d8a-03d8-11e2-86c4-c4461903c065"

        values = {
            'dbuuid': dbuuid,
            'nbr_users': 0,
            'nbr_active_users': 0,
            'nbr_share_users': 0,
            'nbr_active_share_users': 0,
            'dbname': 'stw_dontuse_test',
            'db_create_date': fields.Datetime.to_string(now()),
            'version': '7.0-alpha',
            'contracts': [],
            'language': 'en_US',
            'company_name': 'Stephane Wirtel',
            'company_phone': '+',
            'company_email': 'stw@openerp.com',
        }

        params = {
            'arg0': repr(values),
            'action': 'update',
        }

        query = requests.get(publisher_warranty_url, params=params)

        self.assertEqual(query.status_code, 200)
        self.assertTrue(query.ok)

    def test_external_ping_one_ping_per_day(self):
        DB = self.env['openerp.enterprise.database']

        # publisher_warranty_url = "http://services.openerp.com/publisher-warranty/"
        publisher_warranty_url = self.base_url() + '/publisher-warranty/'
        dbuuid = "21d20d8a-03d8-11e2-86c4-c4461903c064"
        # dbuuid = "21d20d8a-03d8-11e2-86c4-c4461903c065"
        date_now = now()
        values = {
            'dbuuid': dbuuid,
            'nbr_users': 0,
            'nbr_active_users': 0,
            'nbr_share_users': 0,
            'nbr_active_share_users': 0,
            'dbname': 'stw_dontuse_test',
            'db_create_date': fields.Datetime.to_string(date_now),
            'version': '7.0-alpha',
            'contracts': [],
            'language': 'en_US',
            'company_name': 'Stephane Wirtel',
            'company_phone': '+',
            'company_email': 'stw@openerp.com',
        }

        params = {
            'arg0': repr(values),
            'action': 'update',
        }

        # First ping for new db
        query = requests.get(publisher_warranty_url, params=params)
        self.assertEqual(query.status_code, 200)
        database = DB.search([('db_uuid', '=', dbuuid)])
        self.assertEqual(len(database), 1, "The openerp.enterprise.database should exists now")

        # Ping the same day is ignored
        last_ping_saved = database.last_ping_id
        query = requests.get(publisher_warranty_url, params=params)
        database._compute_last_ping()
        self.assertEqual(query.status_code, 200)
        self.assertEqual(database.last_ping_id, last_ping_saved, "We should not have a new ping created")

        # check that no write on dup
        def _write(self, vals, **kwargs):
            raise Exception("Should not write ...", vals)
        with patch.object(self.registry['openerp.enterprise.database'], 'write', _write):
            last_ping_saved = database.last_ping_id
            self.env['openerp.enterprise.ping'].ping(params['arg0'], '127.0.0.1')

        # check futur ping
        with freeze_time(fields.Date.to_string(now() + timedelta(days=5))):
            # First ping in 5 day
            last_ping_saved = database.last_ping_id
            query = requests.get(publisher_warranty_url, params=params)
            database._compute_last_ping()
            self.assertEqual(query.status_code, 200)
            self.assertNotEqual(database.last_ping_id, last_ping_saved, "We should have new ping created")

            # second ping in 5 days, same days that previous, should do noop
            last_ping_saved = database.last_ping_id
            query = requests.get(publisher_warranty_url, params=params)
            database._compute_last_ping()
            self.assertEqual(query.status_code, 200)
            self.assertEqual(database.last_ping_id, last_ping_saved, "We should not have new ping created")

            # Third ping in 5 days, same days that previous but with force_ping
            values['force_ping'] = True
            params['arg0'] = repr(values)
            last_ping_saved = database.last_ping_id
            query = requests.get(publisher_warranty_url, params=params)  # create a new ping
            database._compute_last_ping()
            self.assertEqual(query.status_code, 200)
            self.assertNotEqual(database.last_ping_id, last_ping_saved, "We should have a new ping, because force_ping mode set")


class MailTestCase(EnterpriseDatabaseCommon, HttpCase):
    def setUp(self):
        tools.config['publisher_warranty_url'] = self.base_url() + '/publisher-warranty/'

        super(MailTestCase, self).setUp()

    def test_publisher_contract(self):
        # we store the message in the system
        ping_data = {
            'name': u'Your Company',
            'language': u'en_US',
            'phone': False,
            'nbr_users': 3,
            'nbr_active_share_users': False,
            'dbname': 'trunk',
            'nbr_share_users': False,
            'dbuuid': u'04415138-0bd3-11e2-9d7f-c4461903c065',
            'version': '7.0alpha',
            'email': u'info@yourcompany.com',
            'db_create_date': u'2012-10-01 14:19:31',
            'nbr_active_users': 0
        }

        self.env['openerp.enterprise.ping'].ping(str(ping_data), '127.0.0.1')
